/*
 * Simple concurrent client which repeatedly receives lines from
 * clients and return them prefixed by the string "ACK: ".  Release
 * the connection with a client upon receipt of the string "quit"
 * (case sensitive) or upon a connection close by the client.
 *
 * Of course, the concurrent variant is almost identical to the purely
 * iterative one (i.e., the one that serves one client at a time),
 * except for the call to fork.  Do compare the complexity of this
 * code with the complexity of the code that simulates concurrency.
 *
 * This version also preallocates the processes used by the whole
 * system.  Up to 10 client handling processes can run concurrently.
 *
 * Note: The master process exist after creating the slave processes.
 * However, the slave processes continue to run in the background,
 * don't panic.  This become evident when you do a `ps aux'.
 *
 * Problem is, however that you can no longer stop the server by
 * typing Control-C as you are used.  Instead, use the command
 * `killall cpserv' (replace cpserv with the actual name of your
 * program if necessary).
 *
 * By Stefan Bruda
 */

#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include "tcp-utils.h"

const int max_processes = 10;    // number of preallocated processes.

/*
 * We do not want left certain signals (most notably SIGPIPE)
 * unhandled since that will kill our server, but at the same time we
 * handle the respective events by other means.  The solution is to
 * provide an empty signal handler.
 */
void signal_ignore (int signal) {
    /* NOP */
}

/*
 * Repeatedly receives requests from the client and responds to them.
 * If the received request is an end of file or "quit", terminates the
 * connection.  Note that an empty request also terminates the
 * connection.  Same as for the purely iterative server.
 */
void do_client (int sd) {
    const int ALEN = 256;
    char req[ALEN];
    const char* ack = "ACK: ";
    int n;
    
    // Loop while the client has something to say...
    while ((n = readline(sd,req,ALEN-1)) != recv_nodata) {
        if (strcmp(req,"quit") == 0) {
            printf("Received quit, sending EOF.\n");
            shutdown(sd,1);
            return;
        }
        send(sd,ack,strlen(ack),0);
        send(sd,req,strlen(req),0);
        send(sd,"\n",1,0);
    }
    // read 0 bytes = EOF:
    printf("Connection closed by client.\n");
    shutdown(sd,1);
}

/*
 * Cleans up "zombie children," i.e., processes that did not really
 * die.  Will run each time the parent receives a SIGCHLD signal.
 */
void cleanup_zombies (int sig) {
    int status;
    printf("Cleaning up... \n");
    while (waitpid(-1,&status,WNOHANG) >= 0)
        /* do nothing */;
}

int main (int argc, char** argv) {
    const int port = 9007;   // port to listen to
    const int qlen = 32;     // incoming connections queue length
    
    int msock, ssock = 0;           // master and slave socket
    struct sockaddr_in client_addr; // the address of the client...
    unsigned int client_addr_len = sizeof(client_addr); // ... and its length
    
    msock = passivesocket(port,qlen);
    if (msock < 0) {
        perror("passivesocket");
        return 1;
    }
    printf("Server up and listening on port %d.\n", port);

    signal(SIGCHLD,cleanup_zombies);  // clean up errant children...
                                      // (not really needed, because we preallocate)
    signal(SIGPIPE,signal_ignore);    // SIGPIPE is sent when a client closes connection;
                                      // this is already handled so we ignore the signal.

    // Loop no longer needed!
    // Accept code moved to child process.
  
    for (int i = 0; i < max_processes; i++) { // we create the processes.
    
        int pid = fork();
        if (pid < 0)
            perror("fork");
        else if (pid == 0) {  // child, manages the client-server communication
            while (1) {
                // Accept connection:
                ssock = accept(msock, (struct sockaddr*)&client_addr, &client_addr_len);
                if (ssock < 0) {
                    perror("accept");
                    continue;
                    // we do not return, next time we may be luckier!
                }
                // accept successful
                printf("Incoming client.\n");
                do_client(ssock);
                close(ssock);
                printf("Outgoing client removed.\n");
            }
            return 0; // will never reach this point...
        } // end of child code
        else // parent
            close(ssock);  // parent has no use for ssock, and it goes again
                           // to create the next child.
    }
    return 0;  // processes created, we are done.
}
