/*
 * Simple multithreaded client which repeatedly receives lines from
 * clients and return them prefixed by the string "ACK: ".  Release
 * the connection with a client upon receipt of the string "quit"
 * (case sensitive) or upon a connection close by the client.
 *
 * This version featured a monitor thread, which wakes up each two
 * minutes and reports some information.  The monitoring process is
 * the same as the one used in the textbook.
 *
 * This version also preallocates the threads used by the whole
 * system.  Up to 10 client handling threads can run concurrently.
 *
 * By Stefan Bruda
 */

#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <pthread.h>
#include <stdio.h>
#include <time.h>
#include "tcp-utils.h"

const int max_threads = 10;  // number of preallocated threads.

/*
 * Global variables, for monitoring.  Notice that it is global, and
 * thus visible in all the threads; which is good, because all the
 * client threads will want to write to it so that the monitor can
 * report meaningful information.
 */

struct monitor_t {
    pthread_mutex_t mutex;
    unsigned int con_cur;
    unsigned int con_count;
    unsigned int con_time;
    unsigned int bytecount;
};

/*
 * We do not want left certain signals (most notably SIGPIPE)
 * unhandled since that will kill our server, but at the same time we
 * handle the respective events by other means.  The solution is to
 * provide an empty signal handler.
 */
void signal_ignore (int signal) {
    /* NOP */
}

// Monitor statistics:
monitor_t mon;

/*
 * The function that implements the monitor thread.
 */
void* monitor (void* ignored) {
    const int wakeup_interval = 120; // 2 minutes
    int connections;

    while (1) {
        sleep(wakeup_interval);
        pthread_mutex_lock(&mon.mutex);
        time_t now = time(0);
        if (mon.con_count == 0)
            connections = 1;
        else
            connections = mon.con_count;
        printf("MON: %s\n", ctime(&now));
        printf("MON: currently serving %d clients\n", mon.con_cur);
        printf("MON: average connection time is %d seconds.\n",
               (int)((float)mon.con_time/(float)connections));
        printf("MON: transferred %d bytes per connection on average.\n",
               (int)((float)mon.bytecount/(float)connections));
        printf("MON: (end of information)\n");
        pthread_mutex_unlock(&mon.mutex);
    }
}

/*
 * Repeatedly receives requests from the client and responds to them.
 * If the received request is an end of file or "quit", terminates the
 * connection.  Note that an empty request also terminates the
 * connection.  Same as for the purely iterative or the multi-process
 * server.
 */

void* do_client (int msock) {
    int sd; // slave socket for this thread
    const int ALEN = 256;
    char req[ALEN];
    const char* ack = "ACK: ";
    int n;
    time_t start = time(0);
    struct sockaddr_in client_addr; // the address of the client...
    unsigned int client_addr_len = sizeof(client_addr); // ... and its length

    // Accept connection (moved from main).  When a request arrive, one
    // thread will win and the others will stay blocked; if there are
    // more clients than threads, the extra ones will wait in the TCP
    // queue.
    while (1) {  // we now go into an infinite loop in a thread...
        sd = accept(msock, (struct sockaddr*)&client_addr, &client_addr_len);
        if (sd < 0) {
            if (errno == EINTR) continue;
            perror("accept");
            printf("Cllient discarded!\n");
            continue;
        }
        
        printf("Incoming client...\n");
        // monitor code begins
        pthread_mutex_lock(&mon.mutex);
        mon.con_cur++;
        pthread_mutex_unlock(&mon.mutex);
        // monitor code ends
        
        // Loop while the client has something to say...
        while ((n = readline(sd,req,ALEN-1)) != recv_nodata) {
            if (strcmp(req,"quit") == 0) {
                printf("Received quit, sending EOF.\n");
                break;  // from inner loop, to next client!
            }
            // monitor code begins
            pthread_mutex_lock(&mon.mutex);
            mon.bytecount += n;
            pthread_mutex_unlock(&mon.mutex);
            // monitor code ends
            send(sd,ack,strlen(ack),0);
            send(sd,req,strlen(req),0);
            send(sd,"\n",1,0);
        }
        // read 0 bytes = EOF:
        if (n == recv_nodata) // we could have gotten out because we received "quit"
            printf("Connection closed by client.\n");
        printf("Connection closed, honest!.\n");
        shutdown(sd,1);
        close(sd);
        printf("Outgoing client...\n");
        // monitor code begins
        pthread_mutex_lock(&mon.mutex);
        mon.con_cur--;
        mon.con_count++;
        mon.con_time += time(0) - start;
        pthread_mutex_unlock(&mon.mutex);
        // monitor code ends
    }
    // We will never reach this though...
    return NULL;
}

int main (int argc, char** argv) {
    const int port = 9006;   // port to listen to
    const int qlen = 32;     // incoming connections queue length
    
    // Note that the following are local variable, and thus not shared
    // between threads; especially important for ssock and client_addr.
    
    long int msock; // master socket (slave socket ssock no longer needed!)

    signal(SIGPIPE,signal_ignore);    // SIGPIPE is sent when a client closes connection;
                                      // this is already handled so we ignore the signal.
    
    msock = passivesocket(port,qlen);
    if (msock < 0) {
        perror("passivesocket");
        return 1;
    }
    printf("Server up and listening on port %d.\n", port);

    // Setting up the thread creation:
    pthread_t tt;
    pthread_attr_t ta;
    pthread_attr_init(&ta);
    pthread_attr_setdetachstate(&ta,PTHREAD_CREATE_DETACHED);
    
    // Launch the monitor:
    pthread_create(&tt, &ta, monitor, NULL); 

    // No need for a loop!  We create all the threads at once.  
    // Accept code moved inside the slave threads (i.e., inside
    // do_client).
    
    // Create threads:
    for (int i = 0; i < max_threads; i++)
        if ( pthread_create(&tt, &ta, (void* (*) (void*))do_client, (void*)msock) != 0 ) {
            perror("pthread_create");
            return 1;
        }

    // main thread keeps running (otherwise the other threads will
    // die) but does not really do anything useful.  Note: we could
    // have put the monitor code here, but I just wanted to keep
    // this version as close to mtserv as possible.
    while(1) {
        sleep(1000); // we sleep, so that we do not eat up system resources
    }
    return 0;   // will never reach this anyway...
}
