/*
 * Simple concurrent client which repeatedly receives lines from
 * clients and return them prefixed by the string "ACK: ".  Release
 * the connection with a client upon receipt of the string "quit"
 * (case sensitive) or upon a connection close by the client.
 *
 * This is the "super server" variant, in the sense that it does
 * execve to launch a separate server for each request.  However,
 * there is no dynamic or static configuration, the thing is basically
 * a super server for one type of clients.  True, it does not make
 * that much sense, but it is good enough for didactic purposes.
 *
 * By Stefan Bruda
 */

#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include "tcp-utils.h"

const char* slave_server = "ack";  // name of executable to
                                   // launch for any client...

/*
 * We do not want left certain signals (most notably SIGPIPE)
 * unhandled since that will kill our server, but at the same time we
 * handle the respective events by other means.  The solution is to
 * provide an empty signal handler.
 */
void signal_ignore (int signal) {
    /* NOP */
}

/*
 * Cleans up "zombie children," i.e., processes that did not really
 * die.  Will run each time the parent receives a SIGCHLD signal.
 */
void cleanup_zombies (int sig) {
    int status;
    printf("sserv: cleaning up... \n");
    while (waitpid(-1,&status,WNOHANG) >= 0)
        /* do nothing */;
}

int main (int argc, char** argv, char** envp) {
    const int port = 9007;   // port to listen to
    const int qlen = 32;     // incoming connections queue length

    int msock, ssock;               // master and slave socket
    struct sockaddr_in client_addr; // the address of the client...
    unsigned int client_addr_len = sizeof(client_addr); // ... and its length
  
    msock = passivesocket(port,qlen);
    if (msock < 0) {
        perror("sserv: passivesocket");
        return 1;
    }
    printf("sserv: up and listening on port %d.\n", port);

    signal(SIGCHLD,cleanup_zombies);  // clean up errant children...
    signal(SIGPIPE,signal_ignore);    // SIGPIPE is sent when a client closes connection;
                                      // this is already handled so we ignore the signal.
    
    while (1) {
        // Accept connection:
        ssock = accept(msock, (struct sockaddr*)&client_addr, &client_addr_len);
        if (ssock < 0) {
            perror("sserv: accept");
            return 1;
        }
    
        int pid = fork();
        if (pid < 0)
            perror("sserv: fork");
        else if (pid == 0) {  // child, manages the client-server communication
            close(msock);     // child does not listen to msock...
            printf("sserv: incoming client!\n");
            
            printf("sserv:   closing file descriptor 0... "); 
            fflush(stdout);
            if (close(0)) {  // must do this because dup2 will (justly)
                             // complain otherwise.
                printf("\n");
                perror("sserv:   close");
                return 1;
            }
            else
                printf(" done.\n");

            printf("sserv:   copying slave socket to descriptor 0... "); 
            fflush(stdout);
            if (dup2(ssock,0) < 0) {
                printf("\n");
                perror("sserv:   dup2");
                return 1;
            }
            else
                printf(" done.\n");

            printf("sserv:   passing now the client to the slave server.\n");
            char* slave_server_args[] = {(char*)slave_server, 0};  // no arguments to the slave server...
            execve(slave_server, slave_server_args, envp);
            
            // if we get here execve did not work...
            perror("sserv:   execve");
            shutdown(ssock,1);  //graceful exit
            close(ssock);
            return 1;
        }
        else // parent
            close(ssock);  // parent has no use for ssock...
    }
    return 0;   // will never reach this anyway...
}
